/* ViewMenu.java
 *
 * created: Tue Dec 29 1998
 *
 * This file is part of Artemis
 * 
 * Copyright (C) 1998,1999  Genome Research Limited
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * $Header: /nfs/disk222/yeastpub/Repository/powmap/diana/components/ViewMenu.java,v 1.19 1999/07/16 14:26:31 kmr Exp $
 */

package diana.components;

import Diana;

import diana.*;
import diana.sequence.*;

import uk.ac.sanger.pathogens.*;
import uk.ac.sanger.pathogens.embl.InvalidRelationException;

import java.awt.*;
import java.awt.event.*;
import java.io.File;

/**
 *  A popup menu with viewing commands.
 *
 *  @author Kim Rutherford
 *  @version $Id: ViewMenu.java,v 1.19 1999/07/16 14:26:31 kmr Exp $
 **/

public class ViewMenu extends SelectionMenu {
  /**
   *  Create a new ViewMenu object.
   *  @param frame The Frame that owns this Menu.
   *  @param selection The Selection that the commands in the menu will
   *    operate on.
   *  @param entry_group The EntryGroup object where new features/entries will
   *    be added.
   **/
  public ViewMenu (final Frame frame,
                   final Selection selection,
                   final EntryGroup entry_group) {
    super (frame, "View", selection);

    plot_features_item = new MenuItem ("Show Feature Plots");
    plot_features_item.addActionListener (new ActionListener () {
      public void actionPerformed (ActionEvent event) {
        plotSelectedFeatures ();
      }
    });

    view_feature_item = new MenuItem ("View Selected Features",
                                      new MenuShortcut (KeyEvent.VK_V));
    view_feature_item.addActionListener (new ActionListener () {
      public void actionPerformed (ActionEvent event) {
        viewSelectedFeatures ();
      }
    });

    view_selection_item = new MenuItem ("View Selection");
    view_selection_item.addActionListener (new ActionListener () {
      public void actionPerformed (ActionEvent event) {
        new SelectionViewer (getSelection (), entry_group);
      }
    });

    feature_info_item = new MenuItem ("Show Feature Statistics");
    feature_info_item.addActionListener (new ActionListener () {
      public void actionPerformed (ActionEvent event) {
        viewSelectedFeatureInfo ();
      }
    });

    view_feature_bases_item = new MenuItem ("View Feature Bases");
    view_feature_bases_item.addActionListener (new ActionListener () {
      public void actionPerformed (ActionEvent event) {
        viewSelectedFeatureBases ();
      }
    });

    view_feature_aa_item = new MenuItem ("View Feature Amino Acids");
    view_feature_aa_item.addActionListener (new ActionListener () {
      public void actionPerformed (ActionEvent event) {
        viewSelectedFeatureAminoAcids ();
      }
    });

    overview_item = new MenuItem ("Show Overview",
                                  new MenuShortcut (KeyEvent.VK_O));
    overview_item.addActionListener (new ActionListener () {
      public void actionPerformed (ActionEvent event) {
        new EntryGroupInfoDisplay (entry_group);
      }
    });

    add (view_feature_item);
    add (view_selection_item);

    final ExternalProgramVector external_programs =
      Options.getOptions ().getExternalPrograms ();

    for (int i = 0 ; i < external_programs.size () ; ++i) {
      final String program_name = external_programs.elementAt (i).getName ();

      final String new_menu_name;

      if (Diana.isStandAlone ()) {
        new_menu_name =
          "View " + program_name + " results";
      } else {
        new_menu_name = "Selected Features " + program_name + " results";
      }

      MenuItem view_item = new MenuItem (new_menu_name);

      view_item.addActionListener (new ActionListener () {
        public void actionPerformed (ActionEvent event) {
          viewExternalResults (program_name);
        }
      });

      add (view_item);
    }
    addSeparator ();
    add (overview_item);
    addSeparator ();
    add (view_feature_bases_item);
    add (view_feature_aa_item);
    addSeparator ();
    add (feature_info_item);
    add (plot_features_item);

  }

  /**
   *  If more than this number of features are selected than
   *  viewSelectedFeatures () etc. will ask for confirmation.
   **/
  private static final int MAXIMUM_SELECTED_FEATURES = 10;

  /**
   *  Open a view window for each of the selected features.  The viewer will
   *  listen for feature change events and update itself.
   **/
  private void viewSelectedFeatures () {
    if (!checkForSelectionFeatures (MAXIMUM_SELECTED_FEATURES,
                                    "really view all (>" +
                                    MAXIMUM_SELECTED_FEATURES +
                                    ") selected features?")) {
      return;
    }

    final FeatureVector features_to_view = getSelection ().getAllFeatures ();

    for (int i = 0 ; i < features_to_view.size () ; ++i) {
      final Feature selection_feature = features_to_view.elementAt (i);

      new FeatureViewer (selection_feature);
    }
  }
  
  /**
   *  Open an plot viewing window for each of the selected features.  The plot
   *  viewer will listen for feature change events and update itself.
   **/
  private void plotSelectedFeatures () {
    if (!checkForSelectionFeatures (MAXIMUM_SELECTED_FEATURES,
                                    "really plot all (>" +
                                    MAXIMUM_SELECTED_FEATURES +
                                    ") selected features?")) {
      return;
    }

    final FeatureVector features_to_edit = getSelection ().getAllFeatures ();

    for (int i = 0 ; i < features_to_edit.size () ; ++i) {
      final Feature selection_feature = features_to_edit.elementAt (i);

      new FeaturePlotGroup (selection_feature);
    }
  }

  /**
   *  Show the output file from an external program (like fasta) for the
   *  selected Feature objects.  The name of the file to read stored in a
   *  feature qualifier.  The qualifier used is the program name plus "_file".
   **/
  private void viewExternalResults (String program_name) {
    if (!checkForSelectionFeatures (MAXIMUM_SELECTED_FEATURES,
                                    "really view results from " +
                                    "all (>" + MAXIMUM_SELECTED_FEATURES +
                                    ") selected features?")) {
      return;
    }

    final FeatureVector features_to_view = getSelection ().getAllFeatures ();

    for (int i = 0 ; i < features_to_view.size () ; ++i) {
      final Feature this_feature = features_to_view.elementAt (i);
      
      String file_name;

      try {
        file_name = this_feature.getValueOfQualifier (program_name + "_file");
      } catch (InvalidRelationException e) {
        file_name = null;
      }

      if (file_name == null || file_name.length () == 0) {
        new MessageDialog (getParentFrame (),
                           "Message",
                           "No " + program_name + " results for " +
                           this_feature.getIDString ());

        continue;
      }

      final FileViewer file_viewer =
        new FileViewer (program_name + " results for " +
                        this_feature.getIDString ()); 

      try {
//        System.out.println ("root: " + this_feature.getEntry ().getRootDocument ());

        final Document root_document =
          this_feature.getEntry ().getRootDocument ();

        final Document document;

        if (root_document == null) {
          // try the current directory
          document = new FileDocument (new File (file_name));
        } else {
          document = root_document.append (file_name);
        }

//        System.out.println ("document: " + document);

        file_viewer.readFile (document.getReader ());
      } catch (java.io.IOException e) {
        file_viewer.dispose ();
//        System.out.println (e);
        new MessageDialog (getParentFrame (),
                           "Message",
                           "Could not read " + program_name + " file: " + e);
      }
    }

//    Diana.getClipboard ().setClip (getSelection ());
  }

  /**
   *  Open a FeatureInfo component for each of the selected features.  The
   *  new component will listen for feature change events and update itself.
   **/
  private void viewSelectedFeatureInfo () {
    if (!checkForSelectionFeatures (MAXIMUM_SELECTED_FEATURES,
                                    "really view statistics for " +
                                    "all (>" + MAXIMUM_SELECTED_FEATURES +
                                    ") selected features?")) {
      return;
    }

    final FeatureVector features_to_view = getSelection ().getAllFeatures ();

    for (int i = 0 ; i < features_to_view.size () ; ++i) {
      final Feature selection_feature = features_to_view.elementAt (i);

      new FeatureInfo (selection_feature);
    }
  }

  /**
   *  View the bases of the selected features.  This creates a
   *  FeatureBaseViewer for each feature.
   **/
  private void viewSelectedFeatureBases () {
    if (!checkForSelectionFeatures (MAXIMUM_SELECTED_FEATURES,
                                    "really view bases for " +
                                    "all (>" + MAXIMUM_SELECTED_FEATURES +
                                    ") selected features?")) {
      return;
    }

    final FeatureVector features_to_view = getSelection ().getAllFeatures ();

    for (int i = 0 ; i < features_to_view.size () ; ++i) {
      final Feature this_feature = features_to_view.elementAt (i);

      new FeatureBaseViewer (this_feature);
    }
  }

  /**
   *  View the bases of the selected CDS features.  This creates a
   *  FeatureBaseViewer for each CDS feature.
   **/
  private void viewSelectedFeatureAminoAcids () {
    if (!checkForSelectionFeatures (MAXIMUM_SELECTED_FEATURES,
                                    "really view amino acids for " +
                                    "all (>" + MAXIMUM_SELECTED_FEATURES +
                                    ") selected features?")) {
      return;
    }

    final FeatureVector features_to_view = getSelection ().getAllFeatures ();

    for (int i = 0 ; i < features_to_view.size () ; ++i) {
      final Feature this_feature = features_to_view.elementAt (i);

      new FeatureAminoAcidViewer (this_feature);
    }
  }

  private MenuItem feature_info_item = null;
  private MenuItem plot_features_item = null;
  private MenuItem view_feature_item = null;
  private MenuItem view_selection_item = null;
  private MenuItem view_fasta_item = null;
  private MenuItem view_feature_bases_item = null;
  private MenuItem view_feature_aa_item = null;
  private MenuItem overview_item = null;
}
