/* DocumentEntry.java
 *
 * created: Wed Dec 30 1998
 *
 * This file is part of Artemis
 * 
 * Copyright (C) 1998,1999  Genome Research Limited
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * $Header: /nfs/disk222/yeastpub/Repository/powmap/uk/ac/sanger/pathogens/embl/DocumentEntry.java,v 1.8 1999/07/08 11:56:16 kmr Exp $
 */

package uk.ac.sanger.pathogens.embl;

import uk.ac.sanger.pathogens.*;

import java.util.Vector;
import java.io.Reader;
import java.io.StringReader;
import java.io.BufferedReader;
import java.io.Writer;
import java.io.FileWriter;
import java.io.File;
import java.io.IOException;

/**
 *  This class extends the Entry class with the data for the entry coming from
 *  a Document object.
 *
 *  @author Kim Rutherford
 *  @version $Id: DocumentEntry.java,v 1.8 1999/07/08 11:56:16 kmr Exp $
 **/

public class DocumentEntry extends EMBLObject
    implements Entry {
  /**
   *  Create a new DocumentEntry object associated with the given Document.
   *  @param document This is the file that we will read from.  This is also
   *    used for saving the entry back to the file it came from and to give
   *    the new object a name.
   *  @exception IOException thrown if there is a problem reading the entry -
   *    most likely ReadFormatException.
   *  @exception InvalidRelationException Thrown if this Feature cannot contain
   *    the given Qualifier.
   **/
  public DocumentEntry (Document document)
      throws IOException, InvalidRelationException {
    this.document = document;

    if (! document.readable ()) {
      // this is a new empty DocumentEntry
      return;
    }

    final Reader in_file = document.getReader ();

    LinePushBackReader pushback_reader = new LinePushBackReader (in_file);

    LineGroup new_line_group;
    
    while ((new_line_group =
            LineGroup.readNextLineGroup (this, pushback_reader)) != null) {
      addLineGroup (new_line_group);
      Thread.yield ();
    }

    System.gc();

    in_file.close ();
  }

  /**
   *  Create a new DocumentEntry that will be a copy of the given Entry.
   **/
  public DocumentEntry (final Entry new_entry) {
    final FeatureEnumeration feature_enum = new_entry.features ();

    while (feature_enum.hasMoreFeatures ()) {
      final Feature new_feature = feature_enum.nextFeature ();

      add (new StreamFeature (new_feature));
    }

    final Sequence new_sequence = new_entry.getSequence ();

    if (new_sequence != null) {
      addLineGroup (new EmblStreamSequence (new_sequence));
    }
  }

  /**
   *  Create a new empty DocumentEntry object with no Document associated
   *  with it.  The new DocumentEntry cannot be saved to a file with save ()
   *  unless save (Document) has been called first.  The save (Document)
   *  method will assign a Document.
   **/
  public DocumentEntry () {
    this.document = null;
  }
 
  /**
   *  Returns true if and only if this entry is read only.  For now this
   *  always returns false.
   **/
  public boolean isReadOnly () {
    return false;
  }
  
  /**
   *  Return the text of the EMBL header of this Entry or null if there is no
   *  header.
   **/
  public String getHeaderText () {
    final StringBuffer buffer = new StringBuffer ();

    for (int i = 0 ; i < line_groups.size () ; ++i) {

      final LineGroup current_line_group = line_groups.elementAt (i);

      if (current_line_group instanceof Unknown) {
        buffer.append (current_line_group.toString ());
      }
    }

    if (buffer.length () > 0) {
      return buffer.toString ();
    } else {
      return null;
    }
  }

  /**
   *  Set the header of this Entry to be the given text.
   *  @return true if and only if the header was successfully set.  Not all
   *    Entry objects can change their header, so it is up to the calling
   *    function to check the return value.
   *  @exception IOException thrown if there is a problem reading the header
   *    from the String - most likely ReadFormatException.
   **/
  public boolean setHeaderText (final String new_header)
      throws IOException {

    final Reader reader = new StringReader (new_header);
    
    final LinePushBackReader pushback_reader = new LinePushBackReader (reader);

    final LineGroupVector new_line_groups = new LineGroupVector ();

    LineGroup new_line_group;
    
    try {
      while ((new_line_group =
              LineGroup.readNextLineGroup (this, pushback_reader)) != null) {
        if (new_line_group instanceof Unknown) {
          new_line_groups.addElement (new_line_group);
        } else {
          throw new ReadFormatException ("the header must contain only " +
                                         "header lines");
        }
      }
    } catch (InvalidRelationException e) {
      throw new ReadFormatException ("the header must contain only " +
                                     "header lines");
    }

    // now remove all the Unkown LineGroup objects from this Entry
    for (int i = line_groups.size () - 1 ; i >= 0  ; --i) {

      final LineGroup current_line_group = line_groups.elementAt (i);

      if (current_line_group instanceof Unknown) {
        line_groups.removeElementAt (i);
      }
    }

    // then add the new LineGroup objects
    for (int i = 0 ; i < new_line_groups.size () ; ++i) {
      line_groups.insertElementAt (new_line_groups.elementAt (i), i);
    }
    
    return true;
  }

  /**
   *  Write this Entry to the Document it came from.  This method uses the
   *  current Document reference (as given by getDocument ()).  This method
   *  will throw a NullPointerException if the is no current Document
   *  ie. if getDocument () return null.  Use save (Document) to save and set
   *  the current Document.
   *  @exception IOException thrown if there is a problem saving the entry.
   **/
  public void save ()
      throws IOException {

//    System.out.println ("in DocumentEntry.save ()");
    save (true);
  }

  /**
   *  Write this Entry to the Document it came from.  This method uses the
   *  current Document reference (as given by getDocument ()).  This method
   *  will throw a NullPointerException if the is no current Document
   *  ie. if getDocument () return null.  Use save (Document) to save and set
   *  the current Document.
   *  @param include_diana_extensions If true then the any diana additions to
   *    the embl file format will be included in the output, otherwise they
   *    will be removed.
   *  @exception IOException thrown if there is a problem saving the entry.
   **/
  public void save (final boolean include_diana_extensions)
      throws IOException {

//    System.out.println ("in DocumentEntry.save ()");
    final Writer out_file = getDocument ().getWriter ();
    writeToStream (out_file, include_diana_extensions);
    out_file.close ();
  }

  /**
   *  Write this Entry to the given Document.
   *  @param document This is the file that we will write to.
   *  @param include_diana_extensions If true then the any diana additions to
   *    the embl file format will be included in the output, otherwise they
   *    will be removed.
   *  @exception IOException thrown if there is a problem saving the entry.
   **/
  public void save (final Document document,
                    final boolean include_diana_extensions)
      throws IOException {

//    System.out.println ("in DocumentEntry.save ()");
    final Writer out_file = document.getWriter ();
    writeToStream (out_file, include_diana_extensions);
    out_file.close ();
  }

  /**
   *  Write this Entry to the given stream.  The output will be in EMBL
   *  format (but see the diana_extensions argument).
   *  @param writer The stream to write to.
   *  @param include_diana_extensions If true then the any diana additions to
   *    the embl file format will be included in the output, otherwise they 
   *    will be removed.
   *  @exception IOException thrown if there is a problem writing the entry.
   **/
  protected void writeToStream (final Writer writer,
                                final boolean include_diana_extensions)
      throws IOException {
    
    for (int i = 0 ; i < line_groups.size () ; ++i) {
      final LineGroup current_line_group = line_groups.elementAt (i);

      current_line_group.writeToStream (writer, include_diana_extensions);
   }

    if (line_groups.size () == 1) {
      final LineGroup first_line_group = line_groups.elementAt (0);

      if (first_line_group.getType () == LineGroup.SEQUENCE &&
          ((StreamSequence) first_line_group) instanceof EmblStreamSequence) {
        // don't write out the "//" end of entry marker if this is raw
        // sequence 
        return;
      }

      if (first_line_group.getType () == LineGroup.FEATURE_TABLE) {
        // don't write out the "//" end of entry marker if this the feature
        // table is the only thing in this entry.
        return;
      }
    } else {
      if (line_groups.size () == 2) {
        final LineGroup second_line_group = line_groups.elementAt (1);

        if (second_line_group.getType () == LineGroup.SEQUENCE &&
            ((StreamSequence) second_line_group) instanceof EmblStreamSequence) {
          // don't write out the "//" end of entry marker if this is raw
          // sequence 
          return;
        }
      }
    }

    LineGroup.writeEndOfEMBLEntry (writer);
  }

  /**
   *  Return the File reference that was passed to the constructor or null if
   *  none was passed.
   **/
  public Document getDocument () {
    return document;
  }

  /**
   *  Return a count of the number of Feature obejcts in this Entry.
   **/
  public int getFeatureCount () {
    return getFeatureTable ().getFeatureCount ();
  }

  /**
   *  Add the given Feature to this Entry.
   **/
  public Feature add (Feature new_feature) {
    if (new_feature.getEntry () != null) {
      throw new Error ("internal error - a feature must have one owner");
    }

    if (new_feature instanceof StreamFeature) {
      final StreamFeature stream_feature = (StreamFeature) new_feature;
      getFeatureTable ().add (stream_feature);
      stream_feature.setDocumentEntry (this);
      return new_feature;
    } else {
      // make a StreamFeature copy of the Feature
      final StreamFeature feature_copy = new StreamFeature (new_feature);

      getFeatureTable ().add (feature_copy);
      feature_copy.setDocumentEntry (this);
      return feature_copy;
    }
  }

  /**
   *  Remove the given Feature from this Entry.
   *  @return true if and only if the Feature was in this Entry.
   **/
  public boolean remove (Feature feature) {
    final StreamFeature feature_from_table =
      getFeatureTable ().remove (feature);

    if (feature_from_table == null) {
      return false;
    } else {
      feature_from_table.setDocumentEntry (null);
      return true;
    }
  }

  /**
   *  Return the ith Feature from this Entry.  This Features are returned in a
   *  consistent order, sorted by the first base of each Feature.
   **/
  public Feature getFeatureAtIndex (int i) {
     return getFeatureTable ().getFeatureAtIndex (i);
  }

  /**
   *  Return the index of the given Feature.  This does the reverse of
   *  getFeatureAtIndex ().
   **/
  public int indexOf (final Feature feature) {
    return getFeatureTable ().indexOf (feature);
  }

  /**
   *  Returns true if and only if this Entry contains the given feature.
   **/
  public boolean contains (final Feature feature) {
    return getFeatureTable ().contains (feature);
  }

  /**
   *  Create a new Feature object of an appropriate type in this Entry.
   *  @param key The new feature key
   *  @param location The Location object for the new feature
   *  @param qualifiers The qualifiers for the new feature (can be null if
   *    there are no qualifiers).
   *  @exception InvalidRelationException Thrown if this Feature cannot contain
   *    the given Qualifier.
   **/
  public Feature createFeature (Key key,
                                Location location,
                                QualifierVector qualifiers)
      throws InvalidRelationException, ReadOnlyException {

    final Feature new_feature = new StreamFeature (key, location, qualifiers);

    add (new_feature);
    
    return new_feature;
  }

  /**
   *  Return a vector containing the references of the Feature objects within
   *  the given range.
   *  @param range Return features that overlap this range - ie the start of
   *    the feature is less than or equal to the end of the range and the end
   *    of the feature is greater than or equal to the start of the range.
   *  @return The non-source key features of this feature table the are within
   *    the given range.  The returned object is a copy - changes will not
   *    effect the FeatureTable object itself.
   **/
  public FeatureVector getFeaturesInRange (Range range) {
    return getFeatureTable ().getFeaturesInRange (range);
  }

  /**
   *  Return a vector containing the references of all the Feature objects in
   *  this Entry.
   *  @return The non-source key features of this Entry.  The returned object
   *    is a copy - changes will not effect the Entry object itself.
   **/
  public FeatureVector getAllFeatures () {
    return getFeatureTable ().getAllFeatures ();
  }

  /**
   *  Returns an enumeration of the Feature objects in this DocumentEntry.
   *  The returned Enumeration object will generate all features in this
   *  object in turn. The first item generated is the item at index 0, then
   *  the item at index 1, and so on.
   **/
  public FeatureEnumeration features () {
    return getFeatureTable ().features ();
  }

  /**
   *  Return the Sequence object from this entry or null if it does not
   *  contain one.
   *  @return a Sequence object for this Entry.  the returned object is
   *    not a copy - changes to it will change the Entry object itself
   **/
  public Sequence getSequence () {
    return (Sequence) getLineGroupByType (LineGroup.SEQUENCE);
  }

  /**
   *  Add a new LineGroup object to this Entry.
   *  @param new_line_group A new LineGroup to add.
   **/
  void addLineGroup (final LineGroup new_line_group) {
    line_groups.addElement (new_line_group);
  }

  /**
   *  Return the name of this Entry or null if it has no name.
   **/
  public String getName () {
    if (getDocument () == null || getDocument ().getName () == null) {
      return null;
    } else {
      return getDocument ().getName ();
    }
  }

  /**
   *  Set the name of this Entry - if possible (the return value will let the
   *  caller know).
   *  @return true if and only if the name was successfully set.  Not all
   *    Entry objects can change there name, so it is up to the calling
   *    function to check the return value.
   **/
  public boolean setName (final String name) {
    if (name == null || name.length () == 0) {
      return false;
    }
    document = new FileDocument (new File (name));
    return true;
  }

  /**
   *  Read from a LinePushBackReader into a feature.
   *  @exception InvalidRelationException Thrown if this Feature cannot contain
   *    the given Qualifier.
   **/
  public void setFeatureFromPushbackReader (Feature feature,
                                            LinePushBackReader reader)
      throws IOException, InvalidRelationException {
    if (!remove (feature)) {
      throw new Error ("internal error Feature and Entry don't match");
    }

    // XXX
    ((StreamFeature) feature).readFromStream (this, reader);

    add (feature);
  }

  /**
   *  Return the FeatureTable object from this entry (which will be created if
   *  this object does not already contain one).  This method has package
   *  scope because other package should add, remove and query Features
   *  indirectly through the Entry class.
   *  @return a FeatureTable object for this Entry.  the returned object is
   *    not a copy - changes to it will change the Entry object itself
   **/
  FeatureTable getFeatureTable () {
    final FeatureTable feature_table =
      (FeatureTable) getLineGroupByType (LineGroup.FEATURE_TABLE);

    if (feature_table == null) {
      final FeatureTable new_feature_table = new FeatureTable ();

      if (line_groups.size () > 0 &&
          line_groups.elementAt (line_groups.size () - 1).getType () ==
          LineGroup.SEQUENCE) {
        // there is a SEQUENCE LineGroup so put the feature table just before
        // it
        line_groups.insertElementAt (new_feature_table,
                                     line_groups.size () - 1);
      } else {
        // no SEQUENCE so add the feature table at the end
        line_groups.insertElementAt (new_feature_table, line_groups.size ());
      }

      return new_feature_table;
    } else {
      return feature_table;
    }
  }

  /**
   *  This contains all the lines (stored as LineGroup objects) from the entry
   *  stream that was passed to the constructor.
   **/
  protected LineGroup getLineGroupByType (int line_group_type) {
    for (int i = 0 ; i < line_groups.size () ; ++i) {

      final LineGroup current_line_group = line_groups.elementAt (i);

      switch (line_group_type) {
      case LineGroup.FEATURE_TABLE:
        if (current_line_group instanceof FeatureTable) {
          return current_line_group;
        }
        break;
      case LineGroup.SEQUENCE:
        if (current_line_group instanceof Sequence) {
          return current_line_group;
        }
        break;
      default:
        continue;
      }
    }
    return null;
  }

  /**
   *  This contains all the lines (stored as LineGroup objects) from the entry
   *  stream that was passed to the constructor.
   **/
  protected LineGroupVector line_groups = new LineGroupVector ();
  
  /**
   *  The Document object that was passed to the constructor.  This should be
   *  the document that this DocumentEntry was read from.
   **/
  private Document document;
}


